#include <csignal>
#include <cstring>
#include <sstream>
#include <string>
#include <thread>

#include "common/helper.h"
#include "common/opt_parser.h"
#include "common/sample_common.h"
#include "common/sample_connection_observer.h"
#include "common/sample_local_user_observer.h"
#include "common/sample_event.h"
#include "common/write_csvfile.h"
#include "common/log.h"

#define DEFAULT_CONNECT_TIMEOUT_MS (3000)
#define DEFAULT_MAX_DATASTREAM_WAIT_TIME -1
#define DEFAULT_STATISTICS_COUNT (100)

static uint64_t receive_datastream_T1 = 0;
static uint64_t send_datastream_T2 = 0;
static uint64_t round_count = 0;
static uint64_t sumSpendTime = 0;
static std::unique_ptr<WriteCSVFileHandle> csv_file_handle = nullptr;
static SampleEvent receive_ready_;
static bool error_flag = false;

struct SampleOptions {
  std::string appId;
  std::string channelId;
  std::string userId;
  std::string data_stream_reliable = "true";
  int statistics_count = DEFAULT_STATISTICS_COUNT;
  std::string vos_ip;
};
static int waitUntilReceiveMessage(int waitMs) {
  return receive_ready_.Wait(waitMs);
}

class SampleConnectionObserverHandle : public SampleConnectionObserver {
 public:
  SampleConnectionObserverHandle() = default;
  virtual ~SampleConnectionObserverHandle() = default;

 public:
  // inherit from agora::rtc::IRtcConnectionObserver
  void onStreamMessageError(agora::user_id_t userId, int streamId, int code,
                            int missed, int cached) override;
};

void SampleConnectionObserverHandle::onStreamMessageError(
    agora::user_id_t userId, int streamId, int code, int missed, int cached) {
  printf(
      "*** Receive StreamMessage error, userId = %s, streamId = %d, code = "
      "%d, missed = %d, cached = %d\n",
      userId, streamId, code, missed, cached);
  error_flag = true;
}

class SampleLocalUserObserverHandle : public SampleLocalUserObserver {
public:
  SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user);
  virtual ~SampleLocalUserObserverHandle();
public:
  // inherit from agora::rtc::ILocalUserObserver
  void onStreamMessage(agora::user_id_t userId, int streamId, const char *data,
                        size_t length) override;
};

SampleLocalUserObserverHandle::SampleLocalUserObserverHandle(agora::rtc::ILocalUser* local_user)
  :SampleLocalUserObserver(local_user){}
SampleLocalUserObserverHandle::~SampleLocalUserObserverHandle(){}

void SampleLocalUserObserverHandle::onStreamMessage(agora::user_id_t userId,
                                                     int streamId,
                                                     const char* data,
                                                     size_t length) {
  if (data == nullptr || length == 0) {
    printf("*** Data stream received fail, data = %p, length = %d\n", data,
           static_cast<int>(length));
    error_flag = true;
    return;
  }
  for (int i = 0; i < length; i++) {
    if (data[i] != (i + 1)) {
      printf("*** Data stream received mismatch, data = %p, length = %d\n",
             data, static_cast<int>(length));
      error_flag = true;
      return;
    }
  }
  receive_datastream_T1 = 0;
  receive_datastream_T1 = now_ms_t();
  /*printf("[ %s ]--> Now  Receive a Data Stream Message,  Data stream
     received:" "userId = %s, streamId = %d, length = %d\n",
         getCurrentSystemTimeChrono().c_str(), userId, streamId,
         static_cast<int>(length));*/
  receive_ready_.Set();
}


static bool exitFlag = false;
static void SignalHandler(int sigNo) {
  exitFlag = true;
  receive_ready_.Set();
}

int main(int argc, char* argv[]) {
  SampleOptions options;
  opt_parser optParser;

  optParser.add_long_opt("token", &options.appId,
                         "The token for authentication / must");
  optParser.add_long_opt("channelId", &options.channelId, "Channel Id / must");
  optParser.add_long_opt("userId", &options.userId, "User Id / default is 0");
  optParser.add_long_opt(
      "reliable", &options.data_stream_reliable,
      "data stream transfer reliable (true/false)/ default reliable");
  optParser.add_long_opt("statistics_count", &options.statistics_count,
                         "statistics count per time");
  optParser.add_long_opt("vos_ip", &options.vos_ip,
                         "set the vos ip:port address ");

  if ((argc <= 1) || !optParser.parse_opts(argc, argv)) {
    std::ostringstream strStream;
    optParser.print_usage(argv[0], strStream);
    std::cout << strStream.str() << std::endl;
    return -1;
  }

  if (options.appId.empty()) {
    AG_LOG(ERROR, "Must provide appId!");
    return -1;
  }

  if (options.channelId.empty()) {
    AG_LOG(ERROR, "Must provide channelId!");
    return -1;
  }

  std::signal(SIGQUIT, SignalHandler);
  std::signal(SIGABRT, SignalHandler);
  std::signal(SIGINT, SignalHandler);

  // Create Agora service
  agora::base::IAgoraService* service = createAndInitAgoraService(false, true, true);
  if (!service) {
    AG_LOG(ERROR, "Failed to creating Agora service!");
    return -1;
  }
  // Create Agora connection
  agora::rtc::RtcConnectionConfiguration ccfg;
  ccfg.clientRoleType = agora::rtc::CLIENT_ROLE_BROADCASTER;
  ccfg.channelProfile = agora::CHANNEL_PROFILE_CLOUD_GAMING;
  ccfg.autoSubscribeAudio = false;
  ccfg.autoSubscribeVideo = false;
  ccfg.audioRecvMediaPacket = true;
  ccfg.videoRecvMediaPacket = true;
  agora::agora_refptr<agora::rtc::IRtcConnection> connection =
      service->createRtcConnection(ccfg);
  if (!connection) {
    AG_LOG(ERROR, "Failed to creating Agora connection!");
    return -1;
  }

  // Register connection observer to monitor connection event
  auto connObserver = std::make_shared<SampleConnectionObserverHandle>();
  connection->registerObserver(connObserver.get());

  auto localUserObserver = std::make_shared<SampleLocalUserObserverHandle>(connection->getLocalUser());

  // set vos server ip address
  agora::base::IAgoraParameter* agoraParameter =
      connection->getAgoraParameter();
  if (options.vos_ip.size() > 0) {
    std::string vos_ip_info = "\"" + options.vos_ip + "\"";
    std::string vos_list = "{\"rtc.vos_list\":[" + vos_ip_info + "]}";
    agoraParameter->setParameters(vos_list.c_str());
  }

  // Connect to Agora channel
  if (connection->connect(options.appId.c_str(), options.channelId.c_str(),
                          options.userId.c_str())) {
    AG_LOG(ERROR, "Failed to connect to Agora channel!");
    return -1;
  }
  // Wait until connected before sending media stream
  connObserver->waitUntilConnected(DEFAULT_CONNECT_TIMEOUT_MS);
  std::unique_ptr<WriteCSVFileHandle> csv_handle_(
      new WriteCSVFileHandle("./out_file/receive_datastream_delay.csv"));
  csv_file_handle = std::move(csv_handle_);
  csv_file_handle->open();

  int send_stream_id = 12345;

  if (options.data_stream_reliable == "true" ||
      options.data_stream_reliable == "1") {
    connection->createDataStream(&send_stream_id, true, true, false);
    printf("------------------reliable----------------------------------\n");
  } else {
    connection->createDataStream(&send_stream_id, false, false, false);
    printf("------------------unreliable----------------------------------\n");
  }
  std::string msg = std::string(254, 0);
  int i = 1;
  for (auto iter = msg.begin(); iter != msg.end(); iter++, i++)
    *iter = i;
  while (!exitFlag) {
    error_flag = false;
    int iret = waitUntilReceiveMessage(DEFAULT_MAX_DATASTREAM_WAIT_TIME);
    if (iret < 0 || exitFlag) {
      continue;
    }
    /*printf("[ %s ]--> Now  Send a Data Stream Message,  Data stream Send: "
           "streamId = %d, length = %d\n",
           getCurrentSystemTimeChrono().c_str(), send_stream_id,
           static_cast<int>(msg.length()));*/
    send_datastream_T2 = 0;
    send_datastream_T2 = now_ms_t();

    connection->sendStreamMessage(send_stream_id, msg.data(), msg.length());
    if (!error_flag) {
      int spendTime = send_datastream_T2 - receive_datastream_T1;
      sumSpendTime += spendTime;
      round_count++;
      printf(
          "recv datastream then send datastream round_count:%lu, "
          "spendTime:%d\n",
          round_count, spendTime);
      delayTimeBaseInfo delay_info;
      delay_info.round_count_ = round_count;
      delay_info.spendTime_ = spendTime;
      delay_info.data_length = msg.length();
      csv_file_handle->write_csvfile(delay_info);

      if (round_count % options.statistics_count == 0) {
        double averageSpendTime =
            sumSpendTime / (double)options.statistics_count;
        sumSpendTime = 0;
        printf(
            "the average spendTime from send datastream to receive "
            "datastream per %d times is %f \n",
            options.statistics_count, averageSpendTime);
      }
    } else {
      printf("receive datastream packet have error!!!\n");
    }
  }
  // Unregister connection observer
  connection->unregisterObserver(connObserver.get());
  // Disconnect from Agora channel
  bool disconnected = connection->disconnect();
  if (!disconnected) {
    AG_LOG(ERROR, "Failed to disconnect from Agora channel!");
    return -1;
  }
  AG_LOG(INFO, "Disconnected from Agora channel successfully");
  connObserver.reset();
  localUserObserver.reset();
  connection = nullptr;
  csv_file_handle->close();
  csv_file_handle.reset();
  csv_file_handle = nullptr;
  return 0;
}
